from typing import Self

from selenium.webdriver.remote.webdriver import WebDriver

from .web_view import WebView


class Page(WebView):
    """The class for representing a web page.

    It contains the element and components of the page and utils methods for
    page manipulation.

    """

    APP_ROOT: str

    def __init__(
        self,
        webdriver: WebDriver,
        *,
        app_root: str | None = None,
        wait_timeout: int = 5,
        poll_frequency: float = 0.01,
    ):
        """Initialize page.

        Call `wait_until_loaded` method after initialization.

        Args:
            webdriver: Instance of a class for managing the browser.
            app_root: The URL of base page, by default the value of `APP_ROOT`
                attribute is used.
            wait_timeout: Number of seconds before timing out.
            poll_frequency: Time between checks of `wait` condition, lower
                interval - faster checks. This allows to improve overall tests
                speed.

        """
        super().__init__(
            webdriver,
            app_root=app_root or self.APP_ROOT,
            wait_timeout=wait_timeout,
            poll_frequency=poll_frequency,
        )
        self.wait_until_loaded()

    def check_page_is_loaded(self) -> bool:
        """Return result of check that the page is loaded.

        Some pages can be slow to load and cause problems checking for unloaded
        items. To be sure the page is loaded, this property should return the
        result of checking for the slowest parts of the page.

        """
        return True

    @classmethod
    def open(
        cls,
        webdriver: WebDriver,
        *,
        app_root: str | None = None,
    ) -> Self:
        """Open page and initialize page object.

        Args:
            webdriver: Instance of a WebDriver class for managing the browser.
            app_root: The URL of page, by default the value of `APP_ROOT`
                attribute is used.

        """
        webdriver.get(url=f"{app_root or cls.APP_ROOT}")
        # hack to not specify app_root in each page init method
        kwargs = {}
        if app_root:
            kwargs = {"app_root": app_root}

        # Mypy raise error on unpacking `kwargs`:
        # "Page" has incompatible type "**dict[str, str]"; expected "int"
        # "Page" has incompatible type "**dict[str, str]"; expected "float"
        return cls(webdriver, **kwargs)  # type: ignore

    @classmethod
    def open_from_url(
        cls,
        webdriver: WebDriver,
        *,
        path: str,
        app_root: str | None = None,
        **kwargs,
    ) -> Self:
        """Open page from relative path and initialize page object.

        Add `path` to `app_root` in browser URL.

        Args:
            webdriver: Instance of a WebDriver class for managing the browser.
            app_root: The URL of page, by default the value of `APP_ROOT`
                attribute is used.
            path: Relative URL.
            **kwargs: Additional arguments passed to the
                page object initialization.

        """
        # hack to not specify app_root in each page init method
        if app_root:
            kwargs["app_root"] = app_root

        # We don't use `page.navigate_relative` here because we need to
        # navigate to relative url before page is initialized, since otherwise
        # `wait_until_loaded` method in page `__init__` method might fail.
        webdriver.get(
            url=cls._get_full_relative_url(app_root or cls.APP_ROOT, path),
        )

        page = cls(webdriver, **kwargs)
        return page

    def refresh(self) -> None:
        """Refresh web page and wait until it is loaded."""
        self.webdriver.refresh()
        self.wait_until_loaded()

    def wait_until_loaded(self, timeout: float | None = None) -> None:
        """Wait until page is loaded."""
        wait = self.get_wait(timeout)
        wait.until(
            method=lambda _: self.check_page_is_loaded(),
            message=(
                f"Page `{self.__class__}` didn't loaded in "
                f"{wait._timeout} seconds! Didn't wait for `True` from "
                "`check_page_is_loaded` method."
            ),
        )

    def navigate(self, url: str) -> None:
        """Navigate absolute URL.

        Replace the browser URL with the entered one.

        """
        self.webdriver.get(url)

    def navigate_relative(self, relative_url: str = "/") -> None:
        """Navigate to URL relative to application root.

        Args:
            relative_url (str): Relative URL

        """
        self.webdriver.get(
            self._get_full_relative_url(self.app_root, relative_url),
        )

    def click_on_page(self) -> None:
        """Click on (1, 1) coordinates of page (left upper corner).

        Allows you to move focus away from an element, for example, if it
        is currently unavailable for interaction.

        """
        from selenium.webdriver.common.actions.action_builder import (
            ActionBuilder,
        )

        action = ActionBuilder(self.webdriver)
        action.pointer_action.move_to_location(1, 1).click()
        action.perform()

    @staticmethod
    def _get_full_relative_url(app_root: str, relative_url: str) -> str:
        """Add relative URL to application root URL.

        Args:
            app_root: The URL of page, by default the value of `APP_ROOT`
                attribute is used.
            relative_url (str): Relative URL

        """
        if app_root.endswith("/"):
            # https://www.youtube.com/ -> https://www.youtube.com
            app_root = app_root[:-1]

        if relative_url.startswith("/"):
            # /watch_list -> watch_list
            relative_url = relative_url[1:]

        return f"{app_root}/{relative_url}"



from __future__ import annotations

from typing import TYPE_CHECKING, NoReturn, overload

from pomcorn import locators

if TYPE_CHECKING:
    from pomcorn import WebView, XPathElement


class Element:
    """Descriptor for init `PomcornElement` as attribute by locator.

    .. code-block:: python

        # Example
        from pomcorn import Page, Element

        class MainPage(Page):
            title_element = Element(locators.ClassLocator("page-title"))

    """

    cache_attribute_name = "cached_elements"

    @overload
    def __init__(
        self,
        locator: locators.XPathLocator | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        *,
        relative_locator: locators.XPathLocator | None = None,
    ) -> None: ...

    def __init__(
        self,
        locator: locators.XPathLocator | None = None,
        *,
        relative_locator: locators.XPathLocator | None = None,
    ) -> None:
        """Initialize descriptor.

        Use `relative_locator` if you need to include `base_locator` of
        instance, otherwise use `locator`.

        If descriptor is used for instance of ``Page``, then
        ``relative_locator`` is not needed, since element will be searched
        across the entire page, not within some component.

        """
        self.locator = locator
        self.relative_locator = relative_locator

    def __set_name__(self, _owner: type, name: str) -> None:
        """Save attribute name for which descriptor is created."""
        self.attribute_name = name

    def __get__(
        self,
        instance: WebView | None,
        _type: type[WebView],
    ) -> XPathElement:
        """Get element with stored locator."""
        if not instance:
            raise AttributeError("This descriptor is for instances only!")
        return self.prepare_element(instance)

    def prepare_element(self, instance: WebView) -> XPathElement:
        """Init and cache element in instance.

        Initiate element only once, and then store it in an instance and
        return it each subsequent time. This is to avoid calling
        `wait_until_visible` multiple times in the init of component.

        If the instance doesn't already have an attribute to store cache, it
        will be set.

        If descriptor is used for ``Component`` and
        ``self.is_relative_locator=True``, element will be found by sum of
        ``base_locator`` of that component and passed locator of descriptor.

        """
        if not getattr(instance, self.cache_attribute_name, None):
            setattr(instance, self.cache_attribute_name, {})

        cache = getattr(instance, self.cache_attribute_name, {})
        if cached_element := cache.get(self.attribute_name):
            return cached_element

        element = instance.init_element(
            locator=self._prepare_locator(instance),
        )
        cache[self.attribute_name] = element

        return element

    def _prepare_locator(self, instance: WebView) -> locators.XPathLocator:
        """Prepare a locator by arguments.

        Check that only one locator argument is passed, or none.
        If only `relative_locator` was passed, `base_locator` of instance will
        be added to specified in descriptor arguments. If only `locator` was
        passed, it will return only specified one.

        Raises:
            ValueError: If both arguments were passed or neither or
                ``relative_locator`` used not in ``Component``.

        """
        if self.relative_locator and self.locator:
            raise ValueError(
                "You need to pass only one of the arguments: "
                "`locator` or `relative_locator`.",
            )

        if not self.relative_locator:
            if not self.locator:
                raise ValueError(
                    "You need to pass one of the arguments: "
                    "`locator` or `relative_locator`.",
                )
            return self.locator

        from pomcorn import Component

        if self.relative_locator and isinstance(instance, Component):
            return instance.base_locator // self.relative_locator

        raise ValueError(
            f"`relative_locator` should be used only if descriptor used in "
            f"component. `{instance}` is not a component.",
        )

    def __set__(self, *args, **kwargs) -> NoReturn:
        raise ValueError("You can't reset an element attribute value!")



from pomcorn.locators.base_locators import (
    Locator,
    TInitLocator,
    TLocator,
    XPathLocator,
)
from pomcorn.locators.xpath_locators import (
    ButtonWithTextLocator,
    ClassLocator,
    DataTestIdLocator,
    ElementWithTextLocator,
    IdLocator,
    InputByLabelLocator,
    InputInLabelLocator,
    NameLocator,
    PropertyLocator,
    TagNameLocator,
    TextAreaByLabelLocator,
)

__all__ = (
    "ButtonWithTextLocator",
    "ClassLocator",
    "DataTestIdLocator",
    "ElementWithTextLocator",
    "IdLocator",
    "InputByLabelLocator",
    "InputInLabelLocator",
    "Locator",
    "NameLocator",
    "PropertyLocator",
    "TInitLocator",
    "TLocator",
    "TagNameLocator",
    "TextAreaByLabelLocator",
    "XPathLocator",
)



from .base_locators import XPathLocator


class TagNameLocator(XPathLocator):
    """Locator to look for elements with tag by Xpath."""

    def __init__(self, tag: str):
        """Init XPathLocator.

        Specify the query as the string ``//tag``, where ``tag`` is the name of
        the html tag.

        """
        super().__init__(query=f"//{tag}")


class PropertyLocator(XPathLocator):
    """Locator to look for elements with property by XPath."""

    def __init__(
        self,
        prop: str,
        value: str,
        container: str = "*",
        exact: bool = False,
    ):
        """Init XPathLocator.

        Args:
            prop: The name of the html tag property.
            value: The value of property.
            container: The tag in which the property should be. The default is
                ``*``, which means "any tag".
            exact: Specify whether the value of the property being searched
                must match exactly. By default, the search is based on a
                partial match of the value.

        """
        partial_query = f'//{container}[contains(@{prop}, "{value}")]'
        exact_query = f'//{container}[@{prop}="{value}"]'

        super().__init__(query=exact_query if exact else partial_query)


class DataTestIdLocator(PropertyLocator):
    """Locator to look for elements with custom `testid` property.

    Inherits from ``PropertyLocator`` and sets the default values
    ``prop="data-testid"`` and ``exact=True``.

    """

    def __init__(
        self,
        value: str,
        container: str = "*",
    ):
        """Init XPathLocator.

        Args:
            value: The value of ``testid`` property.
            container: The tag in which the property should be. The default is
                ``*``, which means "any tag".

        """
        super().__init__(
            prop="data-testid",
            value=value,
            container=container,
            exact=True,
        )


class IdLocator(PropertyLocator):
    """Locator to look for elements with ID by Xpath.

    Inherits from ``PropertyLocator`` and sets the default values ``prop="id"``
    and ``exact=True``.

    """

    def __init__(
        self,
        value: str,
        container: str = "*",
    ):
        """Init XPathLocator.

        Args:
            value: The value of ``id`` property.
            container: The tag in which the property should be. The default is
                ``*``, which means "any tag".

        """
        super().__init__(
            prop="id",
            value=value,
            container=container,
            exact=True,
        )


class NameLocator(PropertyLocator):
    """Locator to look for elements with name by Xpath.

    Inherits from ``PropertyLocator`` and sets the default values
    ``prop="name"`` and ``exact=True``.

    """

    def __init__(self, value: str, container: str = "*"):
        """Init XPathLocator.

        Args:
            value: The value of ``name`` property.
            container: The tag in which the property should be. The default is
                ``*``, which means "any tag".

        """
        super().__init__(
            prop="name",
            value=value,
            container=container,
            exact=True,
        )


class ElementWithTextLocator(XPathLocator):
    """Locator to look for elements with text by XPath."""

    def __init__(self, text: str, element: str = "*", exact: bool = False):
        """Init XPathLocator.

        Args:
            text: The text that should be inside the tag.
            element: The tag in which the text should be. The default is
                ``*``, which means "any tag".
            exact: Specify whether the value of the property being searched
                must match exactly. By default, the search is based on a
                partial match of the value.

        """
        exact_query = f"//{element}[./text()={self._escape_quotes(text)}]"
        partial_query = f"//{element}[contains(.,{self._escape_quotes(text)})]"

        super().__init__(query=exact_query if exact else partial_query)


class ClassLocator(PropertyLocator):
    """Locator to look for elements with partial class by XPath.

    Inherits from ``PropertyLocator`` and sets the default value
    ``prop="class"``.

    """

    def __init__(
        self,
        class_name: str,
        container: str = "*",
        exact: bool = False,
    ):
        """Init XPathLocator.

        Args:
            class_name: The name of tag class.
            container: The tag in which the property should be. The default is
                ``*``, which means "any tag".
            exact: Specify whether the value of the property being searched
                must match exactly. By default, the search is based on a
                partial match of the value.

        """
        super().__init__(
            prop="class",
            value=class_name,
            container=container,
            exact=exact,
        )


class ButtonWithTextLocator(ElementWithTextLocator):
    """Locator to looking for button with text by XPath.

    Inherits from ``ElementWithTextLocator`` and sets the default value
    ``element="button"``.

    """

    def __init__(self, text: str, exact: bool = False):
        """Init XPathLocator.

        Args:
            text: The text that should be inside the button tag.
            exact: Specify whether the value of the property being searched
                must match exactly. By default, the search is based on a
                partial match of the value.

        """
        super().__init__(text=text, element="button", exact=exact)


class InputInLabelLocator(XPathLocator):
    """Locator to looking for input with label by XPath.

    Specify the query as the string
    ``//label[contains(., "label")]//input``, where ``label`` is the text of
    the input label.

    .. code-block:: html

        # Example
        <label>Title</label>
            <input value="Value">
        </label>

    """

    def __init__(self, label: str):
        """Init XPathLocator."""
        super().__init__(
            query=f"//label[contains(., {self._escape_quotes(label)})]//input",
        )


class InputByLabelLocator(XPathLocator):
    """Locator to looking for input next to label by XPath.

    Specify the query as the string
    ``//label[contains(., "label")]/following-sibling::input``, where ``label``
    is the text of the input label.

    .. code-block:: html

        # Example
        <div>
            <label for="InputWithLabel">Title</label>
            <input id="InputWithLabel" value="Value">
        </div>

    """

    def __init__(self, label: str):
        """Init XPathLocator."""
        super().__init__(
            query=(
                f"//label[contains(., {self._escape_quotes(label)})]"
                "/following-sibling::input"
            ),
        )


class TextAreaByLabelLocator(XPathLocator):
    """Locator to looking for textarea with label by XPath.

    Specify the query as the string
    ``//*[label[contains(text(), "{label}")]]/textarea``, where ``label``
    is the text of the textarea label.

    """

    def __init__(self, label: str):
        """Init XPathLocator."""
        super().__init__(
            query=(
                "//*[label[contains(text(), "
                f"{self._escape_quotes(label)})]]/textarea"
            ),
        )



from contextlib import contextmanager

from selenium.webdriver import ActionChains
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.wait import WebDriverWait

from pomcorn.element import PomcornElement, XPathElement

from . import locators, waits_conditions
from .locators.base_locators import TInitLocator


class WebView:
    """Class for storing basic shortcuts for interacting with the browser."""

    def __init__(
        self,
        webdriver: WebDriver,
        *,
        app_root: str,
        wait_timeout: float,
        poll_frequency: float = 0,
    ):
        """Initialize webview.

        Args:
            webdriver: Instance of a class for managing the browser.
            app_root: The URL of browser.
            wait_timeout: Number of seconds before timing out.
            poll_frequency: Time between checks of `wait` condition, lower
                interval - faster checks. This allows to improve overall tests
                speed.

        """
        self.webdriver = webdriver
        self.app_root = app_root
        self.wait_timeout = wait_timeout
        self.poll_frequency = poll_frequency
        self.wait = self.get_wait(self.wait_timeout)

    def init_element(
        self,
        locator: TInitLocator,
    ) -> PomcornElement[TInitLocator]:
        """Shortcut for initializing Element instances.

        Note: To be consistent with the method of the same name in
        ``Component``, try to use keyword when specifying the ``locator``
        argument whenever possible.

        Args:
            locator: Instance of a class to locate the element in the browser.

        """
        return PomcornElement(web_view=self, locator=locator)

    def init_elements(
        self,
        locator: locators.XPathLocator,
    ) -> list[XPathElement]:
        """Shortcut for initializing many Element instances via single locator.

        Note: Only supports Xpath locators.

        Note: To be consistent with the method of the same name in
        ``Component``, try to use keyword when specifying the ``locator``
        argument whenever possible.

        Args:
            locator: Instance of a class to locate the element in the browser.

        """
        assert isinstance(
            locator,
            locators.XPathLocator,
        ), "Only supports Xpath locators!"
        elements_count = len(self._get_elements(locator=locator))
        return [
            self.init_element(
                locator=locators.XPathLocator(
                    query=f"({locator.query})[{index + 1}]",
                ),
            )
            for index in range(elements_count)
        ]

    def iter_locators(
        self,
        locator: locators.XPathLocator,
        only_visible: bool = False,
    ) -> list[locators.XPathLocator]:
        """Get the list of the locators where each of them match an element.

        For example, there are multiple elements on the page that matches
        `XPathLocator("//a")`.

        This method return the set of locators like:
        * `XPathLocator("(//a)[1]")`
        * `XPathLocator("(//a)[2]")`

        Args:
            locator: Instance of a class to locate the element in the browser.
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements,
                otherwise all the elements (including not visible) will be
                counted.

        """
        result = []
        elements_count = len(
            self._get_elements(locator=locator, only_visible=only_visible),
        )
        for index in range(1, elements_count + 1):
            # Need to wrap it into parentheses to iterate by index when locator
            # is complex: https://sqa.stackexchange.com/a/39465
            result.append(
                locators.XPathLocator(
                    query=f"({locator.query})[{index}]",
                ),
            )
        return result

    def get_wait(
        self,
        timeout: float | None = None,
    ) -> WebDriverWait[WebDriver]:
        """Get `WebDriverWait` instance.

        If no arguments are provided, returns the default wait instance.

        """
        if not timeout:
            return self.wait

        return WebDriverWait(
            driver=self.webdriver,
            timeout=timeout,
            poll_frequency=self.poll_frequency,
        )

    @property
    def current_url(self) -> str:
        """Return the current webdriver URL."""
        return self.webdriver.current_url

    def _get_element(
        self,
        locator: locators.Locator,
        only_visible: bool = True,
    ) -> WebElement:
        """Get WebElement from page by using locator.

        Args:
            locator: Instance of a class to locate the element in the browser.
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements,
                otherwise all the elements (including not visible) will be
                counted.

        """
        if only_visible:
            self.wait_until_locator_visible(locator=locator)
        return self.webdriver.find_element(*locator)

    def _get_elements(
        self,
        locator: locators.Locator,
        only_visible: bool = False,
    ) -> list[WebElement]:
        """Get WebElements from page by using locator.

        Args:
            locator: Instance of a class to locate the element in the browser.
            only_visible: Flag for viewing visible elements. If this is `True`,
                then this method will only get visible elements, otherwise
                all (default) the elements (including not visible) will be
                counted.

        """
        if only_visible:
            self.wait_until_locator_visible(locator=locator)
        return self.webdriver.find_elements(*locator)

    def wait_until_url_contains(
        self,
        url: str,
        timeout: float | None = None,
    ) -> None:
        """Wait until browser's url contains input url.

        By default, method waits for `self.wait_timeout` seconds.
        If you need to change timeout, you can specify it in `timeout`
        argument.

        Raises:
            TimeoutException: If after `self.wait._timeout` seconds the wait
                has not ended.

        """
        wait = self.get_wait(timeout)
        wait.until(
            method=expected_conditions.url_contains(url),
            message=(
                f"Url doesn't contain `{url}` in {wait._timeout} "
                f"seconds! The current URL is `{self.current_url}`."
            ),
        )

    def wait_until_url_not_contains(
        self,
        url: str,
        timeout: float | None = None,
    ) -> None:
        """Wait until browser's url doesn't not contains input url.

        By default, method waits for `self.wait_timeout` seconds.
        If you need to change timeout, you can specify it in `timeout`
        argument.

        Raises:
            TimeoutException: If after `self.wait._timeout` seconds the wait
                has not ended.

        """
        wait = self.get_wait(timeout)
        wait.until(
            method=waits_conditions.url_not_matches(url),
            message=(
                f"Url does contain `{url}` in {wait._timeout} seconds! "
                f"The current URL is `{self.current_url}`."
            ),
        )

    def wait_until_url_changes(
        self,
        url: str | None = None,
        timeout: float | None = None,
    ) -> None:
        """Wait until url changes.

        Args:
            url: Browser URL which should be changed. If the argument is not
                input, will be used `self.current_url`.
            timeout: Number of seconds to wait until timing out. By default,
                method waits for `self.wait_timeout` seconds.

        Raises:
            TimeoutException: If after `self.wait._timeout` seconds the wait
                has not ended.

        """
        url = url or self.current_url
        wait = self.get_wait(timeout)
        wait.until(
            method=expected_conditions.url_changes(url),
            message=(
                f"Url didn't changed from {url} in {wait._timeout} "
                f"seconds! The current URL is `{self.current_url}`."
            ),
        )

    def wait_until_locator_visible(
        self,
        locator: locators.Locator,
        timeout: float | None = None,
    ) -> None:
        """Wait until element matching locator becomes visible.

        Args:
            locator: Instance of a class to locate the element in the browser.
            timeout: Number of seconds to wait until timing out. By default,
                method waits for `self.wait_timeout` seconds.

        Raises:
            TimeoutException: If after `self.wait._timeout` seconds the wait
                has not ended.

        """
        wait = self.get_wait(timeout)
        wait.until(
            method=expected_conditions.visibility_of_element_located(
                locator=(locator.by, locator.query),
            ),
            message=(
                f"Unable to locate {locator} in {wait._timeout} seconds!"
            ),
        )

    def wait_until_locator_invisible(
        self,
        locator: locators.Locator,
        timeout: float | None = None,
    ) -> None:
        """Wait until element matching locator becomes invisible.

        Args:
            locator: Instance of a class to locate the element in the browser.
            timeout: Number of seconds to wait until timing out. By default,
                method waits for `self.wait_timeout` seconds.

        Raises:
            TimeoutException: If after `self.wait._timeout` seconds the wait
                has not ended.

        """
        wait = self.get_wait(timeout)
        wait.until(
            method=expected_conditions.invisibility_of_element_located(
                locator=(locator.by, locator.query),
            ),
            message=(
                f"{locator} is still visible in {wait._timeout} seconds!"
            ),
        )

    def wait_until_clickable(
        self,
        locator: locators.Locator,
        timeout: float | None = None,
    ) -> None:
        """Wait until element matching locator becomes clickable.

        Args:
            locator: Instance of a class to locate the element in the browser.
            timeout: Number of seconds to wait until timing out. By default,
                method waits for `self.wait_timeout` seconds.

        Raises:
            TimeoutException: If after `self.wait._timeout` seconds the wait
                has not ended.

        """
        wait = self.get_wait(timeout)
        wait.until(
            method=expected_conditions.element_to_be_clickable(
                mark=(locator.by, locator.query),
            ),
            message=(
                f"{locator} isn't clickable after {wait._timeout} seconds!"
            ),
        )

    def wait_until_text_is_in_element(
        self,
        text: str,
        locator: locators.Locator,
        timeout: float | None = None,
    ) -> None:
        """Wait until text is present in the specified element by locator.

        Args:
            locator: Instance of a class to locate the element in the browser.
            text: Text that should be presented in element.
            timeout: Number of seconds to wait until timing out. By default,
                method waits for `self.wait_timeout` seconds.

        Raises:
            TimeoutException: If after `self.wait._timeout` seconds the wait
                has not ended.

        """
        wait = self.get_wait(timeout)
        wait.until(
            method=expected_conditions.text_to_be_present_in_element(
                locator=(locator.by, locator.query),
                text_=text,
            ),
            message=(
                f"{locator} doesn't have `{text}` after {wait._timeout} "
                "seconds!"
            ),
        )

    def wait_until_not_exists_in_dom(
        self,
        element: PomcornElement[locators.TLocator] | locators.TLocator,
        timeout: float | None = None,
    ):
        """Wait until element ceases to exist in DOM.

        Args:
            element: Instance of a class to locate the element in the browser
                or instance of element.
            timeout: Number of seconds to wait until timing out. By default,
                method waits for `self.wait_timeout` seconds.

        Raises:
            TimeoutException: If after `self.wait._timeout` seconds the wait
                has not ended.

        """
        wait = self.get_wait(timeout)
        wait.until(
            method=waits_conditions.element_not_exists_in_dom(element),
            message=(
                f"{element} is still exists in DOM after {wait._timeout} "
                "seconds!"
            ),
        )

    def drag_and_drop(self, source: WebElement, target: WebElement):
        """Perform drag and drop.

        Args:
            source: The web element instance to drag.
            target: The web element instance to drag into.

        """
        ActionChains(self.webdriver).drag_and_drop(source, target).perform()

    def scroll_to(self, target: WebElement):
        """Scroll page to target.

        Scroll to the center of target vertically and to the center of target
        horizontally.

        Args:
            target: The web element instance to scroll to.

        """
        # behavior="instant" - to scroll without animation
        # block="center" - vertical scrolling up to center
        # inline="center"- horizontal scrolling up to center
        script = (
            "arguments[0].scrollIntoView("
            "{behavior: 'instant', block: 'center', inline: 'center'}"
            ");"
        )
        self.webdriver.execute_script(script, target)

    def scroll_to_top(self):
        """Scroll browser to top."""
        self.webdriver.execute_script(
            script="window.scrollBy(0, -document.body.scrollHeight)",
        )

    def scroll_to_bottom(self):
        """Scroll browser to bottom."""
        self.webdriver.execute_script(
            script="window.scrollBy(0, document.body.scrollHeight)",
        )

    def get_input_value(self, label: str) -> str:
        """Find input element by label and get it's value."""
        return self.init_element(
            locator=locators.InputByLabelLocator(label=label),
        ).get_value()

    def execute_javascript(self, script: str, *args):
        """Execute simple javascript.

        Args:
            script: JavaScript code as a string object.
            *args: Any applicable arguments for your JavaScript.

        """
        self.webdriver.execute_script(script, *args)

    def switch_to_default(self):
        """Switch webdriver's focus to default content."""
        self.webdriver.switch_to.default_content()

    def switch_to_iframe(self, locator: locators.Locator):
        """Switch webdriver's focus to iframe.

        Args:
            locator: Instance of a class to locate the element in the browser.

        """
        self.webdriver.switch_to.frame(self._get_element(locator))

    @contextmanager
    def iframe_switcher_manager(self, locator: locators.Locator):
        """Context manager for interacting with iframes.

        Args:
            locator: Instance of a class to locate the element in the browser.

        """
        self.switch_to_iframe(locator=locator)
        yield
        self.switch_to_default()



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Generic

from selenium.common.exceptions import StaleElementReferenceException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.select import Select

from pomcorn import locators

if TYPE_CHECKING:
    from pomcorn.web_view import WebView


class PomcornElement(Generic[locators.TLocator]):
    """The class to represent a simple element (tag) on the page.

    Contains methods for the interaction with an element on the browser page.

    """

    def __init__(self, web_view: WebView, locator: locators.TLocator):
        """Init page element.

        Args:
            web_view: Instance of a webview.
            locator: Instance of a class to locate the element in the browser.

        """
        self.web_view = web_view
        self.locator = locator

    def wait_until_visible(self, timeout: float | None = None):
        """Wait until element becomes visible.

        By default, method waits for `self.web_view.wait_timeout` seconds.
        If you need to change timeout, you can specify it in `timeout`
        argument.

        Raises:
            TimeoutException: If after `self.web_view.wait_timeout` seconds
                the wait has not ended.

        """
        self.web_view.wait_until_locator_visible(
            locator=self.locator,
            timeout=timeout,
        )

    def wait_until_invisible(
        self,
        timeout: float | None = None,
    ):
        """Wait until element becomes invisible.

        By default, method waits for `self.web_view.wait_timeout` seconds.
        If you need to change timeout, you can specify it in `timeout`
        argument.

        Raises:
            TimeoutException: If after `self.web_view.wait_timeout` seconds
                the wait has not ended.

        """
        self.web_view.wait_until_locator_invisible(
            locator=self.locator,
            timeout=timeout,
        )

    def wait_until_clickable(
        self,
        timeout: float | None = None,
    ):
        """Wait until element becomes clickable.

        By default, method waits for `self.web_view.wait_timeout` seconds.
        If you need to change timeout, you can specify it in `timeout`
        argument.

        Raises:
            TimeoutException: If after `self.web_view.wait_timeout` seconds
                the wait has not ended.

        """
        self.web_view.wait_until_clickable(
            locator=self.locator,
            timeout=timeout,
        )

    def wait_until_text_is_in_element(
        self,
        text: str,
        timeout: float | None = None,
    ):
        """Wait until text is present in element.

        By default, method waits for `self.web_view.wait_timeout` seconds.
        If you need to change timeout, you can specify it in `timeout`
        argument.

        Raises:
            TimeoutException: If after `self.web_view.wait_timeout` seconds
                the wait has not ended.

        """
        self.web_view.wait_until_text_is_in_element(
            text=text,
            locator=self.locator,
            timeout=timeout,
        )

    def wait_until_not_exists_in_dom(self, timeout: float | None = None):
        """Wait until element ceases to exist in DOM.

        By default, method waits for `self.web_view.wait_timeout` seconds.
        If you need to change timeout, you can specify it in `timeout`
        argument.

        Raises:
            TimeoutException: If after `self.web_view.wait_timeout` seconds
                the wait has not ended.

        """
        self.web_view.wait_until_not_exists_in_dom(
            element=self.locator,
            timeout=timeout,
        )

    def get_element(self, only_visible: bool = True) -> WebElement:
        """Get selenium instance(WebElement) of element.

        Args:
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements,
                otherwise all the elements (including not visible) will be
                counted.

        """
        return self.web_view._get_element(
            locator=self.locator,
            only_visible=only_visible,
        )

    @property
    def exists_in_dom(self) -> bool:
        """Check if element is present in html, can be not visible."""
        return len(self.web_view._get_elements(locator=self.locator)) != 0

    @property
    def is_displayed(self) -> bool:
        """Check if element is displayed.

        If element is not present in the html, return `False`.

        """
        elements = self.web_view._get_elements(locator=self.locator)
        if not elements:
            return False

        try:
            return elements[0].is_displayed()
        except StaleElementReferenceException:
            # Sometimes an element may disappear before we check its visibility
            return False

    @property
    def is_enabled(self) -> bool:
        """Check if element is enabled.

        It is primarily used with buttons.

        """
        return self.get_element().is_enabled()

    @property
    def is_selected(self) -> bool:
        """Check if element is selected.

        It is predominantly used with radio buttons, dropdowns and checkboxes.

        """
        return self.get_element().is_selected()

    def fill(
        self,
        text: str,
        only_visible: bool = True,
        clear: bool = True,
    ):
        """Fill element with text.

        Args:
            text: The text that will be sent to the element to be filled.
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements,
                otherwise all elements (including not visible) will be counted.
            clear: Whether the element needs to be cleared before filling it
                or not (default `True`).

        """
        if clear:
            self.clear(only_visible=only_visible)
        self.send_keys(str(text), only_visible=only_visible)

    def clear(self, only_visible: bool = True):
        """Clear element (input) and it's value.

        Args:
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements,
                otherwise all the elements (including not visible) will be
                counted.

        """
        cmd_ctrl = Keys.CONTROL
        if sys.platform.lower() == "darwin":
            cmd_ctrl = Keys.COMMAND

        self.send_keys(cmd_ctrl + "a", only_visible=only_visible)
        self.send_keys(Keys.BACK_SPACE, only_visible=only_visible)

    def send_keys(self, keys: str, only_visible: bool = True):
        """Send keys to element.

        Simulate user keystrokes.

        Args:
            keys: The names of the keys in the form of a single string.
                More keys: https://www.selenium.dev/selenium/docs/api/py/webdriver/selenium.webdriver.common.keys.html
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements,
                otherwise all the elements (including not visible) will be
                counted.

        """
        self.get_element(only_visible=only_visible).send_keys(*keys)

    def get_text(self, only_visible: bool = True) -> str:
        """Get text from element.

        Args:
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements,
                otherwise all the elements (including not visible) will be
                counted.

        """
        return self.get_element(only_visible=only_visible).text

    def get_attribute(
        self,
        attribute_name: str,
        only_visible: bool = True,
    ) -> str:
        """Get value of attribute from element.

        If attribute is not found, return empty string.

        Args:
            attribute_name: Element attribute name..
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements,
                otherwise all the elements (including not visible) will be
                counted.

        """
        return (
            self.get_element(only_visible=only_visible).get_attribute(
                name=attribute_name,
            )
            or ""
        )

    def set_attribute(
        self,
        attribute_name: str,
        value: str,
        only_visible: bool = True,
    ):
        """Set value to element attribute.

        Args:
            attribute_name: Element attribute name.
            value: New value for attribute.
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements,
                otherwise all the elements (including not visible) will be
                counted.

        """
        element = self.get_element(only_visible=only_visible)
        self.web_view.execute_javascript(
            f"arguments[0].setAttribute('{attribute_name}',arguments[1])",
            element,
            value,
        )

    def get_value(self, only_visible: bool = True):
        """Get value of `value` attribute from element.

        Args:
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements,
                otherwise all the elements (including not visible) will be
                counted.

        """
        return self.get_attribute(
            attribute_name="value",
            only_visible=only_visible,
        )

    def select(self, value: str, only_visible: bool = True):
        """Perform select on element.

        Args:
            value: Value for selecting an element based on visible text.
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements.

        """
        Select(self.get_element(only_visible)).select_by_visible_text(value)

    def click(
        self,
        only_visible: bool = True,
        wait_until_clickable: bool = True,
    ):
        """Click on element.

        Args:
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements.
            wait_until_clickable: Wait until the element is clickable before
                clicking, or not (default `True`).

        """
        if wait_until_clickable:
            self.wait_until_clickable()
        self.get_element(only_visible=only_visible).click()

    def drag_and_drop(
        self,
        target: PomcornElement[locators.TLocator],
        only_visible: bool = True,
    ):
        """Drag and drop page object on target object.

        Args:
            target: The element instance to drag into.
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements.

        """
        self.web_view.drag_and_drop(
            source=self.get_element(only_visible=only_visible),
            target=target.get_element(only_visible=only_visible),
        )

    def scroll_to(self, only_visible: bool = True):
        """Scroll page until element is visible.

        Args:
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements.

        """
        self.web_view.scroll_to(self.get_element(only_visible=only_visible))

    def hover_to(self, only_visible: bool = True):
        """Hover cursor to element.

        Args:
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements.

        """
        action = ActionChains(self.web_view.webdriver).move_to_element(
            to_element=self.get_element(only_visible=only_visible),
        )
        action.perform()

    def get_value_of_css_property(
        self,
        property_name: str,
        only_visible: bool = True,
    ) -> str:
        """Return value of a CSS property.

        Args:
            property_name: Name of CSS property.
            only_visible: Flag for viewing visible elements. If this is `True`
                (default), then this method will only get visible elements.

        """
        return self.get_element(only_visible).value_of_css_property(
            property_name=property_name,
        )

    def add_debug_mark(self):
        """Set element background to red.

        Should be used only for debugging.

        """
        current_style = self.get_attribute("style")
        self.set_attribute("style", f"{current_style} background: red;")

    def remove_debug_mark(self):
        """Remove debug mark."""
        current_style = self.get_attribute("style")
        self.set_attribute(
            attribute_name="style",
            value=current_style.replace("background: red;", ""),
        )


XPathElement = PomcornElement[locators.XPathLocator]



from pomcorn.component import Component, ListComponent
from pomcorn.descriptors import Element
from pomcorn.element import XPathElement
from pomcorn.page import Page
from pomcorn.web_view import WebView

__all__ = (
    "Component",
    "Element",
    "ListComponent",
    "Page",
    "WebView",
    "XPathElement",
)



import typing
from inspect import isclass
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    get_args,
    get_origin,
    overload,
)

from . import locators
from .element import XPathElement
from .page import Page
from .web_view import WebView

TPage = TypeVar("TPage", bound=Page)


class _EmptyValue:
    """Singleton to use as default value for empty class attribute."""

    def __bool__(self) -> Literal[False]:
        """Allow `EmptyValue` to be used in bool expressions."""
        return False


EmptyValue: Any = _EmptyValue()


class Component(Generic[TPage], WebView):
    """The class to represent a page component that depends on base locator.

    It contains page elements, components and utils methods for page
    manipulation, but as a separate entity that can be reused for different
    pages with common elements.

    Implement wait methods until the component becomes visible or invisible.

    """

    base_locator: locators.XPathLocator

    def __init__(
        self,
        page: TPage,
        base_locator: locators.XPathLocator | None = None,
        wait_until_visible: bool = True,
    ):
        """Initialize component.

        Args:
            page: An instance of the page that uses this component.
            base_locator: Instance of a class to locate the component in the
                browser. Used in relative element initialization methods and
                visibility waits. You also can specify it as attribute.
            wait_until_visible: Whether to wait for the component to become
                visible before completing initialization or not.

        """
        super().__init__(
            page.webdriver,
            app_root=page.app_root,
            wait_timeout=page.wait._timeout,
        )
        self.page = page
        self.base_locator = base_locator or self.base_locator
        self.body = self.init_element(locator=self.base_locator)

        if wait_until_visible:
            self.wait_until_visible()

    @overload
    def init_element(
        self,
        *,
        locator: locators.XPathLocator,
    ) -> XPathElement: ...

    @overload
    def init_element(
        self,
        *,
        relative_locator: locators.XPathLocator,
    ) -> XPathElement: ...

    def init_element(
        self,
        *,
        relative_locator: locators.XPathLocator | None = None,
        locator: locators.XPathLocator | None = None,
    ) -> XPathElement:
        """Initialize element including base locator.

        Use `relative_locator` if you need to include `base_locator`, otherwise
        use `locator`.

        Raises:
            ValueError: If both arguments were passed or neither.

        """
        return super().init_element(
            locator=self._prepare_locator(
                locator=locator,
                relative_locator=relative_locator,
            ),
        )

    @overload
    def init_elements(
        self,
        *,
        locator: locators.XPathLocator | None = None,
    ) -> list[XPathElement]: ...

    @overload
    def init_elements(
        self,
        *,
        relative_locator: locators.XPathLocator | None = None,
    ) -> list[XPathElement]: ...

    def init_elements(
        self,
        *,
        relative_locator: locators.XPathLocator | None = None,
        locator: locators.XPathLocator | None = None,
    ) -> list[XPathElement]:
        """Initialize list of elements including base locator.

        Use `relative_locator` if you need to include `base_locator`, otherwise
        use `locator`.

        Raises:
            ValueError: If both arguments were passed or neither.

        """
        return super().init_elements(
            locator=self._prepare_locator(
                locator=locator,
                relative_locator=relative_locator,
            ),
        )

    def _prepare_locator(
        self,
        *,
        relative_locator: locators.XPathLocator | None = None,
        locator: locators.XPathLocator | None = None,
    ) -> locators.XPathLocator:
        """Prepare a locator by arguments.

        Check that only one locator argument is passed, or none.
        If only `relative_locator` was passed, `base_locator` will be added to
        it. If only `locator` was passed, it will return itself.

        Raises:
            ValueError: If both arguments were passed or neither.

        """
        if relative_locator and locator:
            raise ValueError(
                "You need to pass only one of the arguments: "
                "`locator` or `relative_locator`.",
            )

        if not relative_locator:
            if not locator:
                raise ValueError(
                    "You need to pass one of the arguments: "
                    "`locator` or `relative_locator`.",
                )
            return locator
        return self.base_locator // relative_locator

    def wait_until_visible(self, timeout: float | None = None, **kwargs):
        """Wait until component becomes visible.

        By default, method waits for `self.wait._timeout` seconds.
        If you need to change timeout, you can specify it in `timeout`
        argument.

        """
        self.body.wait_until_visible(timeout)

    def wait_until_invisible(self, timeout: float | None = None, **kwargs):
        """Wait until component becomes invisible.

        By default, method waits for `self.wait._timeout` seconds.
        If you need to change timeout, you can specify it in `timeout`
        argument.

        """
        self.body.wait_until_invisible(timeout)


# Here type ignore added because we can't specify TPage as generic for
# Component, but specifying Page is incorrect
ListItemType = TypeVar("ListItemType", bound=Component)  # type: ignore


class ListComponent(Generic[ListItemType, TPage], Component[TPage]):
    """Class to represent a list-like component.

    It contains standard properties and methods for working with list-like
    components:

    * count
    * all
    * get_item_by_text()

    Waits for `base_item_locator` property  to be overridden or one of the
    attributes (`item_locator` or `relative_item_locator`) to be specified.

    """

    _item_class: type[ListItemType] = EmptyValue

    item_locator: locators.XPathLocator | None = None
    relative_item_locator: locators.XPathLocator | None = None

    def __class_getitem__(cls, item: tuple[type, ...]) -> Any:
        """Create parameterized versions of generic classes.

        This method is called when the class is used as a parameterized type,
        such as MyGeneric[int] or MyGeneric[List[str]].

        We override this method to store values passed in generic parameters.

        Args:
            cls: The generic class itself.
            item: The type used for parameterization.

        Returns:
            type: A parameterized version of the class with the specified type.

        """
        list_cls = super().__class_getitem__(item)  # type: ignore
        cls.__generic_parameters__ = item  # type: ignore
        return list_cls

    def __init__(
        self,
        page: TPage,
        base_locator: locators.XPathLocator | None = None,
        wait_until_visible: bool = True,
    ) -> None:
        # If `_item_class` was not specified in `__init_subclass__`, this means
        # that `ListComponent` is used as a parameterized type
        # (e.g., `List[ItemClass, Page]`).
        if isinstance(self._item_class, _EmptyValue):
            # In this way we check the stored generic parameters and, if first
            # from them is valid, set it as `_item_class`
            first_generic_param = self.__generic_parameters__[0]
            if self.is_valid_item_class(first_generic_param):
                self._item_class = first_generic_param
        super().__init__(page, base_locator, wait_until_visible)

    def __init_subclass__(cls) -> None:
        """Run logic for getting/overriding item_class attr for subclasses."""
        super().__init_subclass__()

        # If class has valid `_item_class` attribute from a parent class
        if cls.is_valid_item_class(cls._item_class):
            # We leave using of parent `item_class`
            return

        # Try to get `item_class` from first generic variable
        list_item_class = cls.get_list_item_class()

        if not list_item_class:
            # If `item_class` is not specified in generic we leave it empty
            # because it maybe not specified in base class but will be
            # specified in child
            return

        cls._item_class = list_item_class

    @property
    def base_item_locator(self) -> locators.XPathLocator:
        """Get the base locator of list item.

        Raises:
            ValueError: If both attributes are specified.
            NotImplementedError: If no attribute has been specified,

        """
        if self.relative_item_locator and self.item_locator:
            raise ValueError(
                "You only need to specify one of the attributes: "
                "`relative_item_locator` - if you want locator nested within "
                "`base_locator`, `item_locator` - otherwise. "
                "Or override `base_item_locator` property.",
            )
        if not self.relative_item_locator:
            if not self.item_locator:
                raise NotImplementedError(
                    "You need to specify one of the arguments: "
                    "`relative_item_locator` - if you want locator nested "
                    "within `base_locator`, `item_locator` - otherwise. "
                    "Or override `base_item_locator` property.",
                )
            return self.item_locator
        return self.base_locator // self.relative_item_locator

    @property
    def count(self) -> int:
        """Get count of list items."""
        return len(self._get_elements(self.base_item_locator))

    @property
    def all(self) -> list[ListItemType]:
        """Get all items of list."""
        # Sometimes `base_item_locator` exists in dom but is not visible
        # and method returns an empty list. That's why we add waiting for this
        if (
            base_item := self.init_element(locator=self.base_item_locator)
        ).exists_in_dom:
            base_item.wait_until_visible()

        items: list[ListItemType] = []
        for locator in self.iter_locators(self.base_item_locator):
            items.append(
                self._item_class(page=self.page, base_locator=locator),
            )
        return items

    @classmethod
    def get_list_item_class(cls) -> type[ListItemType] | None:
        """Return class passed in `Generic[ListItemType]`."""
        base_class = next(
            _class
            for _class in cls.__orig_bases__  # type: ignore
            if isclass(get_origin(_class))
            and issubclass(get_origin(_class), ListComponent)
        )

        # Get first generic variable and return it if it is valid item class
        item_class = get_args(base_class)[0]
        if cls.is_valid_item_class(item_class):
            return item_class

        return None

    @classmethod
    def is_valid_item_class(cls, item_class: Any) -> bool:
        """Check that specified ``item_class`` is valid.

        Valid ``item_class`` should be
        * a class and subclass of ``Component``
        * or TypeAlias based on ``Component``

        """
        if isclass(item_class) and issubclass(item_class, Component):
            return True

        if isinstance(item_class, typing._GenericAlias):  # type: ignore
            type_alias = item_class.__origin__  # type: ignore
            return isclass(type_alias) and issubclass(type_alias, Component)

        return False

    def get_item_by_text(self, text: str, exact: bool = False) -> ListItemType:
        """Get list item by text."""
        locator = self.base_item_locator.contains(
            text=text,
            exact=exact,
        )
        return self._item_class(page=self.page, base_locator=locator)

    def __repr__(self) -> str:
        return (
            "ListComponent("
            f"component={self.__class__}, "
            f"item_class={self._item_class}, "
            f"base_item_locator={self.base_item_locator}, "
            f"count={self.count}, "
            f"items={self.all}, "
            f"page={self.page}"
            ")"
        )

    def __str__(self) -> str:
        return f"{self.all}"



# Custom waits conditions
import re
from collections.abc import Callable

from selenium.common.exceptions import (
    NoSuchElementException,
    StaleElementReferenceException,
)
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.expected_conditions import (
    WebDriverOrWebElement,
)

from pomcorn.locators.base_locators import TLocator

from .element import PomcornElement


def url_not_matches(pattern: str) -> Callable[[WebDriver], bool]:
    """Represent `wait condition` to check that url doesn't match pattern."""

    def check_the_match(driver: WebDriver) -> bool:
        return re.search(pattern, driver.current_url) is None

    return check_the_match


def text_in_element_changes(
    element: PomcornElement[TLocator] | TLocator,
    old_text: str,
) -> Callable[[WebDriver], bool]:
    """Represent `wait condition` to check that text doesn't match old text.

    Args:
        element: The element in which the text will be checked.
        old_text: The old text that should disappear.

    """

    def check_the_match(driver: WebDriver) -> bool:
        target = element

        if isinstance(target, PomcornElement):
            return old_text != target.get_text()

        return old_text != driver.find_element(*target).text

    return check_the_match


def element_not_exists_in_dom(
    element: PomcornElement[TLocator] | TLocator,
) -> Callable[[WebDriver], bool]:
    """Represent `wait condition` to check that element not exists in DOM."""

    def check_the_match(driver: WebDriver) -> bool:
        target = element

        if isinstance(target, PomcornElement):
            return not target.exists_in_dom

        try:
            # Check if driver can find element.
            driver.find_element(*target)
        except (NoSuchElementException, StaleElementReferenceException):
            # In the case of NoSuchElement, returns true because the element is
            # not present in DOM.
            # In the case of StaleElementReference, returns true because stale
            # element reference implies that element is no longer visible.
            return True
        # We return `False` because if `driver` can find element without
        # exception, that means element exists in the DOM.
        return False

    return check_the_match


def text_not_to_be_present_in_element_attribute(
    locator: tuple[str, str],
    attribute_: str,
    text_: str,
) -> Callable[[WebDriverOrWebElement], bool]:
    """Represent `wait condition` that checks that text is not in attribute.

    NOTE: "_" in argument names added for consistent with selenium's built-in
    wait condition `text_to_be_present_in_element_attribute`

    Args:
        locator: A tuple with a By instance and a string value. You can get
            them from the Locators classes.
        attribute_: Attribute name.
        text_: Text that should disappear.

    """

    def check_the_match(driver: WebDriverOrWebElement) -> bool:
        try:
            attribute_text = driver.find_element(*locator).get_attribute(
                name=attribute_,
            )
            if attribute_text is None:
                # If attribute is empty, it means that it contains no text.
                return True
            return text_ not in attribute_text
        except StaleElementReferenceException:
            # If element doesn't exist, it means that its attribute doesn't
            # contain text.
            return True

    return check_the_match



"""Module with `XPathLocator`.

Provide only `XPathLocator` because a locator of this type
is sufficient for all operations and
it also allows to combine locators with `/` operator.
It's better to use one type of locators for consistency.

Example:
  # Search button inside base element
  button_element_locator = base_locator / button_locator

"""

from __future__ import annotations

from collections.abc import Iterator
from typing import Literal, TypeVar

from selenium.webdriver.common.by import By


class Locator:
    """Base locator for looking for elements in page."""

    _ALLOWED_LOCATORS = (
        By.ID,
        By.XPATH,
        By.LINK_TEXT,
        By.PARTIAL_LINK_TEXT,
        By.NAME,
        By.TAG_NAME,
        By.CLASS_NAME,
        By.CSS_SELECTOR,
    )

    def __init__(self, by: str, query: str):
        """Init locator.

        Args:
            by: One of the supported selenium locator strategies.
            query: Query for the strategy.

        """
        if by not in self._ALLOWED_LOCATORS:
            raise ValueError(f"No valid `by` found -> `{by}`")
        self.by = by
        self.query: str = query

    def __iter__(self) -> Iterator[str]:
        """Unpack locator.

        This is necessary because selenium WebDriver methods use
        `By` and `value` tuples to find elements. Thanks to this we can use
        `*locator` for these methods.

        """
        return iter((self.by, self.query))

    def __repr__(self) -> str:
        return f"Locator<By `{self.by}`: Query `{self.query}`>"

    def __str__(self) -> str:
        return self.query


TLocator = TypeVar("TLocator", bound=Locator, covariant=True)

# This need only for element initialization because covariant generic can't be
# used as type for parameter of function
TInitLocator = TypeVar("TInitLocator", bound=Locator)


class XPathLocator(Locator):
    r"""Locator to looking for elements in page by XPath.

    XPathLocator overrides methods for `/` and `//` operators to provide
    "path-like" syntax for locators.

    So we can use `/` and `//` operators to concatenate locators query by `/`
    and `//` accordingly:

        # //\*[@class="class"]

        class_locator = ClassLocator("class")

        # //container[@prop="value"]

        property_locator = PropertyLocator(
            prop="prop",
            value="value",
            container="container",

        )

        # //\*[@class="class"]/container[@prop="value"]

        class_locator / property_locator

        # //\*[@class="class"]//container[@prop="value"]

        class_locator // property_locator

    To extend query of locator you can use `extend_query` method:
        new_locator = class_locator.extend_query("[@some_prop='value']")

        new_locator.query   # //\*[@class="class"][@some_prop="value"]

    All custom locators that inherit `XPathLocator` should be independent and
    start with `//`.

    """

    # We move it to constant to fix flake-8 warning B005:
    # https://pypi.org/project/flake8-bugbear/#:~:text=B005
    divider = "//"

    def __init__(self, query: str):
        """Set related query for locators concatenation.

        Args:
            query: Query for the XPath locator strategy.

        """
        self.related_query = query.lstrip(self.divider)
        super().__init__(by=By.XPATH, query=query)

    def __truediv__(self, other: XPathLocator | str) -> XPathLocator:
        """Override `/` operator to implement following XPath locators.

        "/" used to select the nearest children of the current node.

        """
        return self.prepare_relative_locator(other=other, separator="/")

    def __floordiv__(self, other: XPathLocator | str) -> XPathLocator:
        """Override `//` operator to implement nested XPath locators.

        "//" used to select all descendants (children, grandchildren,
        great-grandchildren, etc.) of current node, regardless of their level
        in hierarchy.

        """
        return self.prepare_relative_locator(other=other, separator="//")

    def __or__(self, other: XPathLocator) -> XPathLocator:
        r"""Override `|` operator to implement variant XPath locators.

        Example:
            span = XPathLocator("//span")
            div = XPathLocator("//div")
            img = XPathLocator("//img")

            (span | div) // img == XPathLocator("(//span | //div)//img")

            span | div // img == XPathLocator("(//span | //div//img)")

        """
        return XPathLocator(query=f"({self.query} | {other.query})")

    def __getitem__(self, value: int | str | XPathLocator) -> XPathLocator:
        """Allow to set xpath expressions or index into the locator query.

        Examples:
            div_locator = XPathLocator("//div") --> `//div`

            # Indexation
            div_locator[0]  --> `(//div)[1]`   # xpath numeration starts with 1
            div_locator[-1] --> `(//div)[last()]`

            # Attribute condition
            div_locator["@type='black'"] --> `//div[@type='black']

            # Checking if there are children
            child_locator = XPathLocator("//label").contains("Schedule")
                --> `//label[contains(., 'Star']`
            div_locator[child_locator] --> `//div[//label[contains(., 'Star']]`

        """
        query = f"({self.query})"

        if isinstance(value, XPathLocator):
            value = value.query

        if isinstance(value, str):
            return XPathLocator(f"{query}[{value}]")

        if value >= 0:
            # `+1` is used here because numeration in xpath starts with 1
            query += f"[{value + 1}]"
        elif value == -1:
            # To avoid ugly locators with `...)[last() - 0]`
            query += "[last()]"
        else:
            query += f"[last() - {abs(value + 1)}]"

        return XPathLocator(query)

    def __bool__(self) -> bool:
        """Return whether query of current locator is empty or not."""
        return bool(self.related_query)

    @classmethod
    def _escape_quotes(cls, text: str) -> str:
        """Escape single and double quotes in given text for use in locators.

        This method is useful when locating elements
        with text containing single or double quotes.

        For example, the text `He's 6'2"` will be transformed into:
        `concat("He", "'", "s 6", "'", "2", '"')`.

        The resulting string can be used in XPath expressions
        like `text()=...` or `contains(.,...)`.

        Returns:
            The escaped text wrapped in `concat()` for XPath compatibility,
            or the original text in double quotes if no escaping is needed.

        """
        if not text or ('"' not in text and "'" not in text):
            return f'"{text}"'

        escaped_parts = []
        buffer = ""  # Temporary storage for normal characters

        for char in text:
            if char not in ('"', "'"):
                buffer += char
                continue
            if buffer:
                escaped_parts.append(f'"{buffer}"')
                buffer = ""
            escaped_parts.append(
                "'" + char + "'" if char == '"' else '"' + char + '"',
            )

        if buffer:
            escaped_parts.append(f'"{buffer}"')

        return f"concat({', '.join(escaped_parts)})"

    def extend_query(self, extra_query: str) -> XPathLocator:
        """Return new XPathLocator with extended query."""
        return XPathLocator(query=self.query + extra_query)

    def contains(self, text: str, exact: bool = False) -> XPathLocator:
        """Return new XPathLocator with search on contained text.

        This is shortcut for the commonly used
        `.extend_query(f"[contains(., '{text}')])`.

        Args:
            text: The text that should be inside the tag.
            exact: Specify whether the text being searched must match exactly.
                By default, the search is based on a partial match.

        """
        partial_query = f"[contains(., {self._escape_quotes(text)})]"
        exact_query = f"[./text()={self._escape_quotes(text)}]"
        return self.extend_query(exact_query if exact else partial_query)

    def prepare_relative_locator(
        self,
        other: XPathLocator | str,
        separator: Literal["/", "//"] = "/",
    ) -> XPathLocator:
        """Prepare relative locator base on queries of two locators.

        If one of parent and other locator queries is empty, the method will
        return only the filled one.

        Args:
            other: Child locator object or str locator query.
            separator: Literal which will placed between locators queries - "/"
                used to select nearest children of current node and "//" used
                to select all descendants (children, grandchildren,
                great-grandchildren, etc.) of current node, regardless of their
                level in hierarchy.

        Raises:
            ValueError: If parent and child locators queries are empty.

        """
        related_query = self.related_query
        if not related_query.startswith("("):
            # Parent query can be bracketed, in which case we don't need to use
            # `//`
            # Example:
            #   (//li)[3] -> valid
            #   //(//li)[3] -> invalid
            related_query = f"//{self.related_query}"

        other = XPathLocator(other) if isinstance(other, str) else other

        locator = XPathLocator(
            query=f"{related_query}{separator}{other.related_query}",
        )

        if self and other:
            return locator

        if not (self or other):
            raise ValueError(
                f"Both of locators have empty query. The `{locator.query}` is "
                "not a valid locator.",
            )

        return self if self else other
